@@ -1,5 +1,5 @@
 /************************************************************************************
-   Copyright (C) 2013,2016 MariaDB Corporation AB
+   Copyright (C) 2013,2019 MariaDB Corporation AB
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -65,7 +65,10 @@ MADB_DsnKey DsnKeys[]=
   {"SSLCRL",         offsetof(MADB_Dsn, SslCrl),            DSN_TYPE_STRING, 0, 0},
   {"SSLCRLPATH",     offsetof(MADB_Dsn, SslCrlPath),        DSN_TYPE_STRING, 0, 0},
   {"SOCKET",         offsetof(MADB_Dsn, Socket),            DSN_TYPE_STRING, 0, 0},
-  {"SAVEFILE",       offsetof(MADB_Dsn, SaveFile),            DSN_TYPE_STRING, 0, 0}, /* 30 */
+  {"SAVEFILE",       offsetof(MADB_Dsn, SaveFile),          DSN_TYPE_STRING, 0, 0}, /* 30 */
+  {"USE_MYCNF",      offsetof(MADB_Dsn, ReadMycnf),         DSN_TYPE_OPTION, MADB_OPT_FLAG_USE_CNF, 0},
+  {"TLSVERSION",     offsetof(MADB_Dsn, TlsVersion),        DSN_TYPE_CBOXGROUP, 0, 0},
+  {"FORCETLS",       offsetof(MADB_Dsn, ForceTls),          DSN_TYPE_BOOL,   0, 0},
   /* Aliases. Here offset is index of aliased key */
   {"SERVERNAME",     DSNKEY_SERVER_INDEX,                   DSN_TYPE_STRING, 0, 1},
   {"USER",           DSNKEY_UID_INDEX,                      DSN_TYPE_STRING, 0, 1},
@@ -76,10 +79,9 @@ MADB_DsnKey DsnKeys[]=
   {NULL, 0, DSN_TYPE_BOOL,0,0}
 };
 
+/* TODO: Shouln't 2nd be DSNKEY_OPTIONS_INDEX? But I seemed to remember it was removed for a reason... */
 #define IS_OPTIONS_BITMAP(key_index) (key_index == DSNKEY_OPTIONS_INDEX || key_index == DSNKEY_OPTIONS_INDEX)
 
-#define GET_FIELD_PTR(DSN, DSNKEY, TYPE) ((TYPE *)((char*)(DSN) + (DSNKEY)->DsnOffset))
-
 typedef struct
 {
   unsigned int Key;
@@ -88,13 +90,22 @@ typedef struct
 } MADB_DsnKeyDep;
 
 /* Define pairs of keys that are switches, i.e. setting one should reset the other.
-   Transitive dependencies have to be defined as dierect dependencies here as well */
+   Transitive dependencies have to be defined as direct dependencies here as well */
 const MADB_DsnKeyDep DsnKeysSwitch[]=
 {
   {DSNKEY_NAMEDPIPE_INDEX, DSNKEY_TCPIP_INDEX,     0},
   {DSNKEY_TCPIP_INDEX,     DSNKEY_NAMEDPIPE_INDEX, 0}
 };
 
+const char TlsVersionName[][8]= {"TLSv1.1", "TLSv1.2", "TLSv1.3"};
+const char TlsVersionBits[]=    {MADB_TLSV11, MADB_TLSV12, MADB_TLSV13};
+
+/* {{{ MADB_DSN_SetDefaults() */
+void MADB_DSN_SetDefaults(MADB_Dsn *Dsn)
+{
+  Dsn->IsTcpIp= 1;
+}
+/* }}} */
 
 /* {{{ MADB_Dsn_Init() */
 MADB_Dsn *MADB_DSN_Init()
@@ -184,6 +195,7 @@ BOOL MADB_DsnSwitchDependents(MADB_Dsn *Dsn, unsigned int Changed)
         break;
       case DSN_TYPE_OPTION:
       case DSN_TYPE_BOOL:
+      case DSN_TYPE_CBOXGROUP:
         {
           KeySet= *GET_FIELD_PTR(Dsn, &DsnKeys[Changed], my_bool);
         }
@@ -239,9 +251,36 @@ my_bool MADB_DsnStoreValue(MADB_Dsn *Dsn, unsigned int DsnKeyIdx, char *Value, m
     }
     break;
   case DSN_TYPE_BOOL:
-    if (*GET_FIELD_PTR(Dsn, DsnKey, my_bool) && OverWrite == FALSE)
-      break;
-    *GET_FIELD_PTR(Dsn, DsnKey, my_bool)= atoi(Value);
+    /* If value is not set or we may overwrite it */
+    if (!(*GET_FIELD_PTR(Dsn, DsnKey, my_bool) && OverWrite == FALSE))
+    {
+      *GET_FIELD_PTR(Dsn, DsnKey, my_bool)= atoi(Value);
+    }
+    break;
+  case DSN_TYPE_CBOXGROUP:
+    /* If value is not set or we may overwrite it */
+    if (! (*GET_FIELD_PTR(Dsn, DsnKey, char) && OverWrite == FALSE))
+    {
+      char IntValue= atoi(Value);
+
+      /* Atm we have only one DSN_TYPE_CBOXGROUP!!!, and following works only for it. If sometime another such field is added,
+         we will need another data structure array, that will bind DSN field with string values and bits for this field.
+         So far we use hardcoded arrays for the singe such field we have atm */
+      if (IntValue == '\0')
+      {
+        unsigned int i;
+
+        IntValue= 0;
+        for (i= 0; i < sizeof(TlsVersionBits); ++i)
+        {
+          if (strcasestr(Value, TlsVersionName[i]) != NULL)
+          {
+            IntValue|= TlsVersionBits[i];
+          }
+        }
+      }
+      *GET_FIELD_PTR(Dsn, DsnKey, char)= IntValue;
+    }
     break;
   case DSN_TYPE_INT:
     if (*GET_FIELD_PTR(Dsn, DsnKey, int) && OverWrite == FALSE)
@@ -290,7 +329,7 @@ my_bool MADB_ReadDSN(MADB_Dsn *Dsn, const char *KeyValue, my_bool OverWrite)
   }
   else 
   {
-    if (Value= strchr(KeyValue, '='))
+    if ((Value= strchr(KeyValue, '=')) != NULL)
     {
       ++Value;
       MADB_RESET(Dsn->DSNName, Value);
@@ -368,9 +407,9 @@ my_bool MADB_SaveDSN(MADB_Dsn *Dsn)
     {
       ret= TRUE;
       /* We do not save DSN_TYPE_OPTION - they are saved as OPTIONS bits */
-      switch(DsnKeys[i].Type){
+      switch (DsnKeys[i].Type) {
       case DSN_TYPE_BOOL:
-          ret= SQLWritePrivateProfileString(Dsn->DSNName, DsnKeys[i].DsnKey, 
+        ret= SQLWritePrivateProfileString(Dsn->DSNName, DsnKeys[i].DsnKey, 
           *GET_FIELD_PTR(Dsn, &DsnKeys[i], my_bool) ? "1" : "0", "ODBC.INI");
         break;
       case DSN_TYPE_INT:
@@ -379,6 +418,12 @@ my_bool MADB_SaveDSN(MADB_Dsn *Dsn)
           ret= SQLWritePrivateProfileString(Dsn->DSNName, DsnKeys[i].DsnKey, Value, "ODBC.INI");
         }
         break;
+      case DSN_TYPE_CBOXGROUP:
+        {
+          _snprintf(Value, 32, "%hu", (short)*GET_FIELD_PTR(Dsn, &DsnKeys[i], char));
+          ret= SQLWritePrivateProfileString(Dsn->DSNName, DsnKeys[i].DsnKey, Value, "ODBC.INI");
+        }
+        break;
       case DSN_TYPE_STRING:
       case DSN_TYPE_COMBO:
         {
@@ -386,8 +431,11 @@ my_bool MADB_SaveDSN(MADB_Dsn *Dsn)
           if (Val && Val[0])
             ret= SQLWritePrivateProfileString(Dsn->DSNName, DsnKeys[i].DsnKey, Val, "ODBC.INI");
         }
+      default:
+        /* To avoid warning with some compilers */
         break;
-      }
+      }  /* switch */
+
       if (!ret)
       {
         SQLInstallerError(1,&ErrNum, Dsn->ErrorMsg, SQL_MAX_MESSAGE_LENGTH, NULL);
@@ -407,6 +455,25 @@ my_bool MADB_SaveDSN(MADB_Dsn *Dsn)
 }
 /* }}} */
 
+
+size_t ConnStringLength(const char * String, char Delimiter)
+{
+  size_t      result= strlen(String);
+  const char *p=      String + result + 1;
+
+  if (Delimiter != '\0')
+  {
+    return result;
+  }
+  /* else - we have string with null terminated key=value pairs with additional NULL after last pair */
+  while (*p)
+  {
+    p+= strlen(p) + 1;
+  }
+
+  return p - String /* Length without ending NULL */;
+}
+
 /* {{{ MADB_ParseConnString */
 my_bool MADB_ParseConnString(MADB_Dsn *Dsn, const char *String, size_t Length, char Delimiter)
 {
@@ -417,9 +484,12 @@ my_bool MADB_ParseConnString(MADB_Dsn *Dsn, const char *String, size_t Length, c
     return FALSE;
 
   if (Length == SQL_NTS)
-    Length= strlen(String);
+  {
+    Length= ConnStringLength(String, Delimiter);
+  }
 
-  Buffer= _strdup(String);
+  Buffer= MADB_ALLOC(Length + 1);
+  Buffer= memcpy(Buffer, String, Length + 1);
   Key=    Buffer;
 
   while (Key && Key < ((char *)Buffer + Length))
@@ -458,8 +528,10 @@ my_bool MADB_ParseConnString(MADB_Dsn *Dsn, const char *String, size_t Length, c
             special= TRUE;
           }
         }
-        else if ((p= strchr(Value, ';')))
+        else if ((p= strchr(Value, Delimiter)))
+        {
           *p= 0;
+        }
         Value= trim(Value);
 
         /* Overwriting here - if an option repeated more than once in the string, its last entrance will determine the value */
@@ -471,13 +543,17 @@ my_bool MADB_ParseConnString(MADB_Dsn *Dsn, const char *String, size_t Length, c
         }
 
         if (p)
-          *p= (special) ? ' ' : ';';
+        {
+          *p= (special) ? ' ' : Delimiter;
+        }
         break;
       }
       ++i;
     }
-    if ((Key= strchr(Value, ';')))
+    if ((Key= strchr(Value, Delimiter)))
+    {
       ++Key;
+    }
   }
   MADB_FREE(Buffer);
   return TRUE;
@@ -490,7 +566,7 @@ my_bool MADB_ParseConnString(MADB_Dsn *Dsn, const char *String, size_t Length, c
 BOOL MADB_ReadConnString(MADB_Dsn *Dsn, const char *String, size_t Length, char Delimiter)
 {
   /* Basically at this point we need DSN name only */
-  if (!MADB_ParseConnString(Dsn, String, Length, ';'))
+  if (!MADB_ParseConnString(Dsn, String, Length, Delimiter))
   {
     return FALSE;
   }
@@ -502,7 +578,7 @@ BOOL MADB_ReadConnString(MADB_Dsn *Dsn, const char *String, size_t Length, char
     MADB_ReadDSN(Dsn, NULL, FALSE);
     /* This redundancy is needed to be able to reset options set in the DSN, e.g. if DSN has Reconnect option selected, and
        connection string has AUTO_RECONNECT=0. Connection string should have precedence */
-    MADB_ParseConnString(Dsn, String, Length, ';');
+    MADB_ParseConnString(Dsn, String, Length, Delimiter);
   }
   return TRUE;
 }
@@ -541,7 +617,7 @@ SQLULEN MADB_DsnToString(MADB_Dsn *Dsn, char *OutString, SQLULEN OutLength)
       case DSN_TYPE_INT:
         if (*GET_FIELD_PTR(Dsn, &DsnKeys[i], int))
         {
-          _snprintf(IntVal, sizeof(IntVal), "%d",*(int *)((char *)Dsn + DsnKeys[i].DsnOffset));
+          _snprintf(IntVal, sizeof(IntVal), "%d", *GET_FIELD_PTR(Dsn, &DsnKeys[i], int));
           Value= IntVal;
         }
         break;
@@ -550,6 +626,15 @@ SQLULEN MADB_DsnToString(MADB_Dsn *Dsn, char *OutString, SQLULEN OutLength)
         {
           Value= "1";
         }
+      default:
+        /* To avoid warning with some compilers */
+        break;
+      case DSN_TYPE_CBOXGROUP:
+        if (*GET_FIELD_PTR(Dsn, &DsnKeys[i], char))
+        {
+          _snprintf(IntVal, sizeof(IntVal), "%hu", (short)*GET_FIELD_PTR(Dsn, &DsnKeys[i], char));
+          Value= IntVal;
+        }
         break;
       }
     }
@@ -565,7 +650,9 @@ SQLULEN MADB_DsnToString(MADB_Dsn *Dsn, char *OutString, SQLULEN OutLength)
   }
 
   if (OutLength && OutString)
-    strncpy_s(OutString, OutLength, TmpStr,  TotalLength);
+  {
+    strncpy_s(OutString, OutLength, TmpStr, TotalLength);
+  }
   return TotalLength;
 }
 /* }}} */
